import argparse

from .audio import DEFAULT_AUDIO_FORMAT
from .video import DEFAULT_SCRAMBLE
from .sanzan import Sanzan, DEFAULT_MODE


def main(args=None):
    parser = argparse.ArgumentParser(description="Sanzan")

    parser.add_argument("input", type=str)

    modes = parser.add_mutually_exclusive_group(required=True)

    modes.add_argument("-e", "--encrypt", action="store_true")
    modes.add_argument("-d", "--decrypt", action="store_true")

    parser.add_argument("-k", "--key", type=str)
    parser.add_argument("-o", "--output", type=str)

    parser.add_argument("-m", "--mode", type=str, default=DEFAULT_MODE)
    parser.add_argument("-a", "--audio_format", type=str, default=DEFAULT_AUDIO_FORMAT)
    parser.add_argument("-s", "--scramble", type=str, default=DEFAULT_SCRAMBLE)

    parser.add_argument("-l", "--light", action="store_true")
    parser.add_argument("-dn", "--denoise", action="store_true")
    parser.add_argument("-f", "--per_frame", action="store_true")
    parser.add_argument("-p", "--preview", action="store_true")
    parser.add_argument("-q", "--quiet", action="store_true")

    parser.add_argument("-pp", "--padding", action="store_false", help="Disable audio padding")

    args = parser.parse_args()

    s = Sanzan(args.input, args.mode)
    if args.key:
        s.set_password(args.key)
    if args.encrypt:
        s.encrypt(args.output, args.scramble, args.preview, args.per_frame, args.quiet, args.padding, args.audio_format, args.denoise, args.light)
    elif args.decrypt:
        s.decrypt(args.output, args.scramble, args.preview, args.per_frame, args.quiet, args.padding, args.audio_format, args.denoise, args.light)


if __name__ == "__main__":
    main()



import numpy as np


class Keygen:
    def __init__(self, seed=None) -> None:
        self.rng = None
        self.seed = seed
        if seed is not None:
            self.seed = bytearray(self.seed, "utf8")

        self.reseed()

    def reseed(self):
        self.rng = np.random.RandomState(self.seed)

    def generate_key(self, length, reseed):
        shuf_order = np.arange(length)
        if reseed:
            self.reseed()
        self.rng.shuffle(shuf_order)
        if reseed:
            self.reseed()

        return shuf_order

    def _generate_rev_key(self, shuf_order):
        unshuf_order_x = np.zeros_like(shuf_order)
        unshuf_order_x[shuf_order] = np.arange(len(shuf_order))

        return unshuf_order_x

    def generate_rev_key(self, length, reseed):
        shuf_order = self.generate_key(length, reseed)

        return self._generate_rev_key(shuf_order)



from yt_dlp import YoutubeDL

YDL_OPTS_DEFAULT = {
    "format": "bestvideo",
    "skip_download": True,
    # "quiet" : True,
    # Skip manifest files
    "extractor_args": {"youtube": {"skip": ["dash", "hls"]}},
}


def get_direct_url(source_url, audio=False):
    """
    Get direct url to best video or audio source via yt_dlp
    """

    ydl_opts = YDL_OPTS_DEFAULT

    if audio:
        ydl_opts["format"] = "bestaudio"

    with YoutubeDL(ydl_opts) as ydl:
        extracted_data = ydl.extract_info(source_url, download=False)

        direct_url = extracted_data.get("url")
        if not direct_url:
            raise Exception(f"Could not find direct url for {source_url}")

        return direct_url



import os
import ffmpeg
from pathlib import Path
from multiprocessing import Process

from .youtube import get_direct_url
from .audio import Audio, DEFAULT_AUDIO_FORMAT
from .video import Video, DEFAULT_SCRAMBLE
from .exceptions import SanzanException

DEFAULT_MODE = "full"


class Sanzan:
    def __init__(self, input_raw, mode=DEFAULT_MODE) -> None:
        self.video = None
        self.audio = None
        self.youtube = False

        video_e = None
        audio_e = None

        input_video = input_raw
        input_audio = input_raw

        if input_raw.startswith("http") and "youtu" in input_raw:
            # Convert youtube url to direct urls
            self.youtube = True

        if mode != "audio":
            try:
                if self.youtube:
                    input_video = get_direct_url(input_raw)
                self.video = Video(input_video)
            except SanzanException as e:
                video_e = e

        if mode != "video":
            try:
                if self.youtube:
                    input_audio = get_direct_url(input_raw, audio=True)
                self.audio = Audio(input_audio)
            except SanzanException as e:
                audio_e = e

        if self.video is None and self.audio is None:
            if video_e is not None:
                raise video_e
            if audio_e is not None:
                raise audio_e

        if self.video is None:
            print(f"Will not handle video")
        if self.audio is None:
            print(f"Will not handle audio")

    def set_password(self, password):
        if self.video:
            self.video.set_password(password)
        if self.audio:
            self.audio.set_password(password)

    def merge(self, video_path, audio_path, output_path):
        print(f"Merging `{video_path}` and `{audio_path}` to `{output_path}`")
        input_video = ffmpeg.input(video_path)
        input_audio = ffmpeg.input(audio_path)

        if Path(output_path).suffix.lower() == "mkv":
            f = ffmpeg.output(input_video, input_audio, output_path, acodec="copy", vcodec="copy")
        else:
            f = ffmpeg.concat(input_video, input_audio, v=1, a=1).output(output_path)

        f.run(quiet=True, overwrite_output=True)

    def _crypt(self, crypt_mode, output, scramble, preview, per_frame, quiet, padding, audio_out_format, denoise, light):
        procs = []

        if self.audio:
            audio_crypt_fn = self.audio.encrypt if crypt_mode == 0 else self.audio.decrypt
            output_audio = f"{output}.{audio_out_format}" if output and self.video else output
            # audio_crypt_fn(output_audio, padding, audio_out_format, denoise)
            procs.append(
                Process(
                    target=audio_crypt_fn,
                    args=(
                        output_audio,
                        padding,
                        audio_out_format,
                        denoise,
                        preview,
                        light,
                    ),
                )
            )

        [p.start() for p in procs]

        if self.video:
            video_crypt_fn = self.video.encrypt if crypt_mode == 0 else self.video.decrypt
            output_video = f"tmp_{output}" if output and self.audio else output
            video_crypt_fn(output_video, scramble, preview, per_frame, quiet)

        [p.join() for p in procs]

        if output and self.video and self.audio:
            self.merge(output_video, output_audio, output)
            os.remove(output_video)
            os.remove(output_audio)

    def encrypt(self, output=None, scramble=DEFAULT_SCRAMBLE, preview=False, per_frame=False, quiet=False, padding=True, audio_out_format=DEFAULT_AUDIO_FORMAT, denoise=False, light=False):
        self._crypt(0, output, scramble, preview, per_frame, quiet, padding, audio_out_format, denoise, light)

    def decrypt(self, output=None, scramble=DEFAULT_SCRAMBLE, preview=False, per_frame=False, quiet=False, padding=True, audio_out_format=DEFAULT_AUDIO_FORMAT, denoise=True, light=False):
        self._crypt(1, output, scramble, preview, per_frame, quiet, padding, audio_out_format, denoise, light)



import os
from numba import njit
import time
from tqdm import tqdm
import cv2

from .keygen import Keygen
from .exceptions import SanzanException

DEFAULT_SCRAMBLE = "rows"


@njit
def _shuffle_rows(frame, order_x):
    return frame[:, order_x]


def get_cap_properties(cap):
    props = {}
    props["width"] = cap.get(cv2.CAP_PROP_FRAME_WIDTH)
    props["height"] = cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
    props["frames"] = cap.get(cv2.CAP_PROP_FRAME_COUNT)
    props["fps"] = cap.get(cv2.CAP_PROP_FPS)

    return props


class Video:
    def __init__(self, input_raw) -> None:
        self.t = time.time()
        self.keygen = None
        self.writer = None
        self.input_file = input_raw

        self.cap = cv2.VideoCapture(self.input_file)
        if not self.cap.isOpened():
            raise SanzanException("Invalid video input path or file format specified")

        self.props = get_cap_properties(self.cap)

        if self.props["frames"] < 0 or self.props["fps"] > 1000:
            raise SanzanException(f"Input does not contain a video source!")

    def set_password(self, password):
        self.keygen = Keygen(password)

    def _generate_keys(self, reseed):
        shuf_order_x = self.keygen.generate_key(int(self.props["width"]), reseed)
        shuf_order_y = self.keygen.generate_key(int(self.props["height"]), reseed)

        return shuf_order_x, shuf_order_y

    def _generate_rev_keys(self, reseed):
        unshuf_order_x = self.keygen.generate_rev_key(int(self.props["width"]), reseed)
        unshuf_order_y = self.keygen.generate_rev_key(int(self.props["height"]), reseed)

        return unshuf_order_x, unshuf_order_y

    def _set_output(self, output):
        # Init VideoWriter object
        self.writer = cv2.VideoWriter(
            output,
            cv2.VideoWriter_fourcc(*"mp4v"),
            self.props["fps"],
            (int(self.props["width"]), int(self.props["height"])),
            isColor=True,
        )
        if not self.writer.isOpened():
            os.remove(output)
            raise Exception("Invalid output path or file format specified")

    def _crypt(self, scramble, preview, per_frame, silent, generate_keys_fn):
        order_x = None
        order_y = None

        for i in tqdm(range(int(self.props["frames"])), desc="[video]", disable=silent, position=0):
            success, frame = self.cap.read()
            if not success:
                print(f"Failed to read frame {i}!")
                break

            if per_frame or order_x is None:
                # Generate new keys every frame if `per_frame` is True
                # Otherwise, use the same keys every frame
                order_x, order_y = generate_keys_fn(not per_frame)

            # Shuffle along y-axis
            if scramble != "cols":
                frame = frame[order_y]

            # Shuffle along x-axis
            if scramble != "rows":
                frame = _shuffle_rows(frame, order_x)

            if self.writer:
                self.writer.write(frame)

            if preview:
                cv2.imshow(self.input_file, frame)
                cv2.waitKey(1)

        self.cap.release()
        if self.writer:
            self.writer.release()
        cv2.destroyAllWindows()

    def encrypt(self, output=None, scramble=DEFAULT_SCRAMBLE, preview=False, per_frame=False, silent=False):
        if self.keygen is None:
            # TODO: Export keys
            self.keygen = Keygen()
        if output:
            self._set_output(output)

        self._crypt(scramble, preview, per_frame, silent, self._generate_keys)

    def decrypt(self, output=None, scramble=DEFAULT_SCRAMBLE, preview=False, per_frame=False, silent=False):
        if self.keygen is None:
            raise Exception(f"No key to decrypt with!")
        if output:
            self._set_output(output)

        self._crypt(scramble, preview, per_frame, silent, self._generate_rev_keys)



from setuptools import setup


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="Sanzan",
    version="2.0.3",
    packages=["sanzan"],
    description="Video encryption while maintaining playability.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "numpy",
        "yt-dlp",
        "opencv-python",
        "tqdm",
        "pydub",
        "numba",
        "requests",
        "ffmpeg-python",
        "pyaudio",
    ],
    entry_points={
        "console_scripts": [
            "sz = sanzan.sz:main",
        ]
    },
)



import ffmpeg
import os
import io
import requests
import pydub
from pydub.playback import play
import numpy as np
from math import floor

from .keygen import Keygen
from .exceptions import SanzanException

DEFAULT_AUDIO_FORMAT = "wav"


class Audio:
    def __init__(self, input_raw) -> None:
        """
        Input can be file path or URL
        """

        self.keygen = None
        self.audio = None
        self.array = None
        self.input_raw = input_raw

        mediainfo = pydub.utils.mediainfo(self.input_raw)
        if not mediainfo.get("sample_rate"):
            raise SanzanException(f"Input does not contain an audio source!")

    def _read(self, input_file):
        audio = pydub.AudioSegment.from_file(input_file)
        array = np.array(audio.get_array_of_samples())

        # Convert to 2-d array for multichannel
        if audio.channels > 1:
            array = array.reshape((-1, audio.channels))

        return audio, array

    def _load(self):
        if self.audio is not None and self.array is not None:
            return

        if self.input_raw.startswith("http"):
            # Retrieve url as bytes
            print(f"\n[audio] Fetching stream `{self.input_raw}`")
            self.input_raw = io.BytesIO(requests.get(self.input_raw).content)

        print(f"\n[audio] Loading `{self.input_raw}`")
        self.audio, self.array = self._read(self.input_raw)
        print(f"\n[audio] Loaded {len(self.array)} samples @ {self.audio.frame_rate}hz")

    def _write(self, output, sr, sw, array, out_format, denoise, preview):
        # Derive channels from array shape
        channels = array.ndim if (array.ndim > 1 and array.shape[1] > 1) else 1

        if sw == 1:
            y = np.int8(array)
        elif sw == 2:
            y = np.int16(array)
        elif sw == 4:
            y = np.int32(array)

        audio_out = pydub.AudioSegment(y.tobytes(), frame_rate=sr, sample_width=sw, channels=channels)

        if not output:
            if preview:
                play(audio_out)
            return

        raw_output = f"tmp_{output}" if output and denoise else output
        audio_out.export(raw_output, format=out_format)

        if denoise:
            ffmpeg.input(raw_output).filter("afftdn", nr="max", nf="max", nt="w").output(output).run(quiet=True, overwrite_output=True)
            os.remove(raw_output)
        
        if preview:
            play(pydub.AudioSegment.from_file(output))

    def set_password(self, password):
        self.keygen = Keygen(password)

    def encrypt(self, output, padding=True, out_format=DEFAULT_AUDIO_FORMAT, denoise=False, preview=False, light=False):
        """
        Padding will extend the array to the nearest sample rate multiple before shuffling
        so that the length can be derived when decrypting, but introduces trailing junk data.
        Padding must be applied to be resistant to length changes from encoding and compression.
        Lossless decryption can be performed if `padding` is False and `out_format` is lossless.
        """

        if self.keygen is None:
            # TODO: Export key as file
            self.keygen = Keygen()

        self._load()

        target_array = self.array
        target_shape = self.array.shape

        if padding:
            padded_dur = floor(self.audio.duration_seconds) + 1
            # Round up to nearest sample rate multiple to get length of padded array
            padded_length = padded_dur * self.audio.frame_rate
            print(f"\n[audio] Padding to {padded_length} samples ({padded_dur}s)")
            target_shape = (padded_length, self.audio.channels) if self.audio.channels > 1 else padded_length
            target_array = np.resize(self.array, target_shape)

        if light:
            dim_3 = (self.audio.channels,) if self.audio.channels > 1 else ()
            target_array = np.reshape(target_array, (self.audio.frame_rate,-1,) + dim_3)

        shuf_order = self.keygen.generate_key(len(target_array), True)
        shuffled_array = target_array[shuf_order]

        if light:
            shuffled_array = np.reshape(shuffled_array, target_shape)

        self._write(output, self.audio.frame_rate, self.audio.sample_width, shuffled_array, out_format, denoise, preview)

    def decrypt(self, output, padding=True, out_format=DEFAULT_AUDIO_FORMAT, denoise=True, preview=False, light=False):
        """
        The array length used during encryption can be derived by rounding down to the nearest sample rate multiple.
        Lossless decryption can be performed if `padding` is False and `out_format` is lossless during encryption.
        """

        if self.keygen is None:
            raise Exception(f"No key to decrypt with!")

        self._load()

        target_array = self.array
        target_shape = self.array.shape

        if padding:
            # Round down to nearest sample rate multiple
            padded_dur = floor(self.audio.duration_seconds)
            padded_length = padded_dur * self.audio.frame_rate
            print(f"\n[audio] Rounding down to {padded_length} samples ({padded_dur}s)")
            target_shape = (padded_length, self.audio.channels) if self.audio.channels > 1 else padded_length
            target_array = np.resize(self.array, target_shape)

        if light:
            dim_3 = (self.audio.channels,) if self.audio.channels > 1 else ()
            target_array = np.reshape(target_array, (self.audio.frame_rate,-1,) + dim_3)

        unshuf_order = self.keygen.generate_rev_key(len(target_array), True)
        unshuffled_array = target_array[unshuf_order]

        if light:
            unshuffled_array = np.reshape(unshuffled_array, target_shape)

        self._write(output, self.audio.frame_rate, self.audio.sample_width, unshuffled_array, out_format, denoise, preview)
